@@ -17,6 +17,7 @@ import pandas as pd
 import wget
 
 from sklearn.metrics import mean_squared_error, cohen_kappa_score, mean_absolute_error, log_loss, roc_auc_score, f1_score
+from sklearn.ensemble import VotingClassifier, VotingRegressor
 import matplotlib.pyplot as plt
 import seaborn as sns
 
@@ -33,7 +34,7 @@ import time
 class hgboost:
     """Create a class hgboost that is instantiated with the desired method."""
 
-    def __init__(self, max_eval=250, threshold=0.5, cv=5, test_size=0.2, val_size=0.2, top_cv_evals=10, random_state=None, verbose=3):
+    def __init__(self, max_eval=250, threshold=0.5, cv=5, test_size=0.2, val_size=0.2, top_cv_evals=10, random_state=None, n_jobs=-1, verbose=3):
         """Initialize hgboost with user-defined parameters.
 
         Parameters
@@ -52,6 +53,9 @@ class hgboost:
             Setup the validation set. This part is kept entirely seperate from the test-size.
         random_state : int, (default : None)
             Fix the random state for validation set and test set. Note that is not used for the crossvalidation.
+        n_jobs : int, (default : -1)
+            The number of jobs to run in parallel for fit. None means 1 unless in a joblib.parallel_backend context.
+            -1 means using all processors.
         verbose : int, (default : 3)
             Print progress to screen.
             0: None, 1: ERROR, 2: WARN, 3: INFO, 4: DEBUG, 5: TRACE
@@ -80,10 +84,11 @@ class hgboost:
         self.algo=tpe.suggest
         self.cv = cv
         self.random_state = random_state
+        self.n_jobs = n_jobs
         self.verbose = verbose
 
     def _fit(self, X, y, pos_label=None):
-        """Learn best performing model.
+        """Fit the best performing model.
 
         Description
         -----------
@@ -139,8 +144,8 @@ class hgboost:
         self.space = params
         # Fit model
         self.results = self._fit(X, y, pos_label=self.pos_label)
-        # Return
-        return self.results
+        # Fin
+        if self.verbose>=3: print('[hgboost] >Fin!')
 
     def _regression(self, X, y, eval_metric, greater_is_better, params):
         # Gather for method, the default metric and greater is better.
@@ -150,6 +155,8 @@ class hgboost:
         self.space = params
         # Fit model
         self.results = self._fit(X, y)
+        # Fin
+        if self.verbose>=3: print('[hgboost] >Fin!')
 
     def xgboost_reg(self, X, y, eval_metric='rmse', greater_is_better=False, params='default'):
         """Xgboost Regression with parameter hyperoptimization.
@@ -260,7 +267,7 @@ class hgboost:
         return self.results
 
     def xgboost(self, X, y, pos_label=None, method='xgb_clf', eval_metric=None, greater_is_better=None, params='default'):
-        """Catboost Classification with parameter hyperoptimization.
+        """Xgboost Classification with parameter hyperoptimization.
 
         Parameters
         ----------
@@ -374,6 +381,122 @@ class hgboost:
         # Return
         return self.results
 
+    def ensemble(self, X, y, pos_label=None, methods=['xgb_clf', 'ctb_clf', 'lgb_clf'], eval_metric=None, greater_is_better=None, voting='soft'):
+        """Ensemble Classification with parameter hyperoptimization.
+
+        Description
+        -----------
+        Fit best model for xgboost, catboost and lightboost, and then combine the individual models to a new one.
+
+        Parameters
+        ----------
+        X : pd.DataFrame
+            Input dataset.
+        y : array-like
+            Response variable.
+        pos_label : string/int.
+            Fit the model on the pos_label that that is in [y].
+        methods : list of strings, (default : ['xgb_clf','ctb_clf','lgb_clf']).
+            The models included for the ensemble classifier or regressor. The clf and reg models can not be combined.
+            * ['xgb_clf','ctb_clf','lgb_clf']
+            * ['xgb_reg','ctb_reg','lgb_reg']
+        eval_metric : str, (default : 'auc')
+            Evaluation metric for the regressor of classification model.
+            * 'auc' : area under ROC curve (two-class classification : default)
+        greater_is_better : bool (default : True)
+            If a loss, the output of the python function is negated by the scorer object, conforming to the cross validation convention that scorers return higher values for better models.
+            * auc :  True -> two-class
+        voting : str, (default : 'soft')
+            Combining classifier using a voting scheme.
+            * 'hard' : using predicted classes.
+            * 'soft' : using the Probabilities.
+
+        Returns
+        -------
+        results : dict
+            * best_params: Best performing parameters.
+            * summary: Summary of the models with the loss and other variables.
+            * model: Ensemble of the best performing models.
+            * val_results: Results on indepedent validation dataset.
+        """
+
+        import copy
+        # Store parameters in object
+        self.results = {}
+        self.voting = voting
+        self.methods = methods
+
+        if np.all(list(map(lambda x: 'clf' in x, methods))):
+            if self.verbose>=3: print('[hgboost] >Create ensemble classification model..')
+            self.method = 'ensemble_clf'
+        elif np.all(list(map(lambda x: 'reg' in x, methods))):
+            if self.verbose>=3: print('[hgboost] >Create ensemble regression model..')
+            self.method = 'ensemble_reg'
+        else:
+            raise ValueError('[hgboost] >Error: The input [methods] must be of type "_clf" or "_reg" but can not be combined.')
+
+        # Check input data
+        X, y, self.pos_label = _check_input(X, y, pos_label, self.method, verbose=self.verbose)
+        # Gather for method, the default metric and greater is better.
+        self.eval_metric, self.greater_is_better = _check_eval_metric(self.method, eval_metric, greater_is_better)
+        # Store the clean initialization in hgb
+        hgb = copy.copy(self)
+
+        # Create independent validation set.
+        if self.method == 'ensemble_clf':
+            X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=self.val_size, random_state=self.random_state, shuffle=True, stratify=y)
+        else:
+            X_train, X_val, y_train, y_val = train_test_split(X, y, test_size=self.val_size, random_state=self.random_state, shuffle=True)
+
+        # Hyperparameter optimization for boosting models
+        models = []
+        for method in methods:
+            # Make copy of clean init
+            hgbM = copy.copy(hgb)
+            hgbM.method = method
+            hgbM._classification(X_train, y_train, eval_metric, greater_is_better, 'default')
+            # Store
+            models.append((method, copy.copy(hgbM.model)))
+            self.results[method] = {}
+            self.results[method]['model'] = copy.copy(hgbM)
+
+        # Create the ensemble model
+        if self.verbose>=3: print('[hgboost] >Fit ensemble model with [%s] voting..' %(self.voting))
+        if self.method == 'ensemble_clf':
+            model = VotingClassifier(models, voting=voting, n_jobs=self.n_jobs, verbose=False)
+            model.fit(X, y==pos_label)
+        else:
+            model = VotingRegressor(models, n_jobs=self.n_jobs)
+            model.fit(X, y)
+        # Store ensemble model
+        self.model = model
+
+        # Validation error for the ensemble model
+        if self.verbose>=3: print('[hgboost] >Evalute [ensemble] model on independent validation dataset (%.0f samples, %.2g%%)' %(len(y_val), self.val_size * 100))
+        # Evaluate results on the same validation set
+        val_score, val_results = self._eval(X_val, y_val, model, verbose=2)
+        if self.verbose>=3: print('[hgboost] >[Ensemble] [%s]: %.4g on independent validation dataset' %(self.eval_metric, val_score['loss']))
+
+        # Validate each of the independent methods to show differences in loss-scoring
+        if self.val_size is not None:
+            self.X_val = X_val
+            self.y_val = y_val
+            for method in methods:
+                # Evaluation
+                val_score_M, val_results_M = self._eval(X_val, y_val, self.results[method]['model'].model, verbose=2)
+                # Store
+                self.results[method]['loss'] = val_score_M['loss']
+                self.results[method]['val_results'] = val_results_M
+                if self.verbose>=3: print('[hgboost] >[%s]  [%s]: %.4g on independent validation dataset' %(method, self.eval_metric, val_score_M['loss']))
+
+        # Store
+        self.results['val_results'] = val_results
+        self.results['model'] = model
+        # self.results['summary'] = pd.concat([hgbX.results['summary'], hgbC.results['summary'], hgbL.results['summary']])
+
+        # Return
+        return self.results
+
     def _set_validation_set(self, X, y):
         """Set the validation set.
 
@@ -443,9 +566,9 @@ class hgboost:
         # Validation error
         val_results = None
         if self.val_size is not None:
-            if self.verbose>=3: print('[hgboost] >Evalute best [%s] model on independent validation dataset (%.0f samples, %.2f%%).' %(self.method, len(self.y_val), self.val_size * 100))
+            if self.verbose>=3: print('[hgboost] >Evalute best [%s] model on independent validation dataset (%.0f samples, %.2g%%)' %(self.method, len(self.y_val), self.val_size * 100))
             # Evaluate results
-            val_score, val_results = self._eval(self.X_val, self.y_val, model, self.space, verbose=2)
+            val_score, val_results = self._eval(self.X_val, self.y_val, model, verbose=2)
             if self.verbose>=3: print('[hgboost] >[%s] on independent validation dataset: %.4g' %(self.eval_metric, val_score['loss']))
 
         # Remove the model column
@@ -513,21 +636,20 @@ class hgboost:
         eval_set = [(self.X_train, self.y_train), (self.X_test, self.y_test)]
         # Make fit with stopping-rule to avoid overfitting. Directly perform evaluation with the eval_set.
         model.fit(self.X_train, self.y_train, eval_set=eval_set, **space['fit_params'])
-        # auc_mean = cross_val_score(model, self.X, self.y, cv=self.cv)
         # Evaluate results
-        out, eval_results = self._eval(self.X_test, self.y_test, model, space, verbose=verbose)
+        out, eval_results = self._eval(self.X_test, self.y_test, model, verbose=verbose)
         # Return
         # The model.best_score is the default eval_metric from eval_set in the fit function. The default depends on the selected method.
         # if self.verbose>=4: print("[hgboost] >best score: {0}, best iteration: {1}".format(model.best_score, model.best_iteration))
         return out, eval_results
 
     def xgb_reg(self, space):
-        reg = xgb.XGBRegressor(**space['model_params'])
+        reg = xgb.XGBRegressor(**space['model_params'], n_jobs=self.n_jobs)
         out, _ = self._train_model(reg, space)
         return out
 
     def lgb_reg(self, space):
-        reg = lgb.LGBMRegressor(**space['model_params'])
+        reg = lgb.LGBMRegressor(**space['model_params'], n_jobs=self.n_jobs)
         out, _ = self._train_model(reg, space)
         return out
 
@@ -537,7 +659,7 @@ class hgboost:
         return out
 
     def xgb_clf(self, space):
-        clf = xgb.XGBClassifier(**space['model_params'])
+        clf = xgb.XGBClassifier(**space['model_params'], n_jobs=self.n_jobs)
         out, _ = self._train_model(clf, space)
         return out
 
@@ -547,12 +669,12 @@ class hgboost:
         return out
 
     def lgb_clf(self, space):
-        clf = lgb.LGBMClassifier(**space['model_params'])
+        clf = lgb.LGBMClassifier(**space['model_params'], n_jobs=self.n_jobs)
         out, _ = self._train_model(clf, space)
         return out
 
     def xgb_clf_multi(self, space):
-        clf = xgb.XGBClassifier(**space['model_params'])
+        clf = xgb.XGBClassifier(**space['model_params'], n_jobs=self.n_jobs)
         out, _ = self._train_model(clf, space)
         return out
 
@@ -619,7 +741,7 @@ class hgboost:
         # Return
         return y_pred, y_proba
 
-    def _eval(self, X_test, y_test, model, space, verbose=3):
+    def _eval(self, X_test, y_test, model, verbose=3):
         """Classifier Evaluation.
 
         Description
@@ -668,7 +790,7 @@ class hgboost:
                 elif self.eval_metric=='f1':
                     loss = results[self.eval_metric]
                 elif self.eval_metric=='auc_cv':
-                    loss = np.mean(cross_val_score(model, self.X_train, self.y_train, cv=self.cv))
+                    loss = np.mean(cross_val_score(model, self.X_train, self.y_train, cv=self.cv, n_jobs=self.n_jobs))
                 else:
                     raise ValueError('[hgboost] >Error: [%s] is not a valid [eval_metric] for [%s].' %(self.eval_metric, self.method))
 
@@ -769,6 +891,9 @@ class hgboost:
         ax : object
 
         """
+        if ('ensemble' in self.method):
+            if self.verbose>=2: print('[hgboost] >Warning: No plot for ensemble is possible yet. <return>')
+            return None
         if not hasattr(self, 'model'):
             print('[hgboost] >No model found. Hint: fit a model first using xgboost, catboost or lightboost <return>')
             return None
@@ -792,6 +917,10 @@ class hgboost:
             Figure axis.
 
         """
+        if ('ensemble' in self.method):
+            if self.verbose>=2: print('[hgboost] >Warning: No plot for ensemble is possible yet. <return>')
+            return None
+
         print('[hgboost] >%.0d-fold crossvalidation is performed with [%s]' %(self.cv, self.method))
         disable = (True if (self.verbose==0 or self.verbose>3) else False)
         ax = None
@@ -803,7 +932,7 @@ class hgboost:
             if ('_clf' in self.method) and not ('_multi' in self.method):
                 _, X_test, _, y_test = train_test_split(self.X, self.y, test_size=self.test_size, random_state=None, shuffle=True, stratify=self.y)
                 # Evaluate model
-                _, cl_results = self._eval(X_test, y_test, self.model, self.space, verbose=0)
+                _, cl_results = self._eval(X_test, y_test, self.model, verbose=0)
                 cv_results[name] = cl_results
             elif '_reg' in self.method:
                 _, X_test, _, y_test = train_test_split(self.X, self.y, test_size=self.test_size, random_state=None, shuffle=True)
@@ -839,13 +968,13 @@ class hgboost:
             Figure axis.
 
         """
+        ax = None
         if not hasattr(self, 'model'):
             print('[hgboost] >No model found. Hint: fit a model first using xgboost, catboost or lightboost <return>')
             return None
         if self.val_size is None:
             print('[hgboost] >No validation set found. Hint: use the parameter [val_size=0.2] first <return>')
             return None
-        ax = None
 
         title = 'Results on independent validation set'
         if ('_clf' in self.method) and not ('_multi' in self.method):
@@ -890,6 +1019,10 @@ class hgboost:
             Figure axis.
 
         """
+        if ('ensemble' in self.method):
+            if self.verbose>=2: print('[hgboost] >Warning: No plot for ensemble is possible yet. <return>')
+            return None, None
+
         top_n = np.minimum(top_n, self.results['summary'].shape[0])
         # getcolors = colourmap.generate(top_n, cmap='Reds_r')
         ascending = False if self.greater_is_better else True
@@ -996,6 +1129,54 @@ class hgboost:
         if return_ax:
             return ax, ax2
 
+    # def plot(self, ylim=None, figsize=(15, 10), return_ax=False):
+    #     """Plot the summary results.
+
+    #     Parameters
+    #     ----------
+    #     ylim : tuple
+    #         Set the y-limit. In case of auc it can be: (0.5, 1)
+    #     figsize: tuple, default (25,25)
+    #         Figure size, (height, width)
+
+    #     Returns
+    #     -------
+    #     ax : object
+    #         Figure axis.
+
+    #     """
+    #     ax1, ax2 = None, None
+    #     if (not hasattr(self, 'model')):
+    #         print('[hgboost] >No model found. Hint: fit a model first using xgboost, catboost, lightboost or ensemble <return>')
+    #         return ax1, ax2
+
+    #     if hasattr(self.model, 'val_result'):
+    #         _, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)
+    #     else:
+    #         _, ax1 = plt.subplots(1, 1, figsize=figsize)
+
+    #     # Plot the ensemble or the other models
+    #     if ('ensemble' in self.method):
+    #         ax1, ax2 = self.plot_summary_ensemble(ylim, figsize, ax1, ax2)
+    #     else:
+    #         ax1, ax2 = plot_summary(self, ylim=ylim, figsize=figsize, return_ax=return_ax, ax1=ax1, ax2=ax2)
+
+    #     return ax1, ax2
+
+    def plot_ensemble(self, ylim, figsize, ax1, ax2):
+        # Get models
+        keys = np.array([*self.results.keys()])
+        if self.method=='ensemble_reg':
+            Iloc = list(map(lambda x: '_reg' in x, keys))
+        else:
+            Iloc = list(map(lambda x: '_clf' in x, keys))
+        methods = keys[Iloc]
+
+        for method in methods:
+            model = self.results[method]
+            # self.results['summary'] = pd.concat([hgbX.results['summary'], hgbC.results['summary'], hgbL.results['summary']])
+            # ax1, ax2 = plot_summary(model, ylim=ylim, figsize=figsize, return_ax=True, method=method, ax1=ax1, ax2=ax2)
+
     def plot(self, ylim=None, figsize=(15, 10), return_ax=False):
         """Plot the summary results.
 
@@ -1012,15 +1193,20 @@ class hgboost:
             Figure axis.
 
         """
-        if not hasattr(self, 'model'):
-            print('[hgboost] >No model found. Hint: fit a model first using xgboost, catboost or lightboost <return>')
-            return None
         ax1, ax2 = None, None
+        if ('ensemble' in self.method):
+            if self.verbose>=2: print('[hgboost] >Warning: No plot for ensemble is possible yet. <return>')
+            self.plot_ensemble(ylim, figsize, ax1, ax2)
+            return ax1, ax2
+        if (not hasattr(self, 'model')):
+            print('[hgboost] >No model found. Hint: fit a model first using xgboost, catboost or lightboost <return>')
+            return ax1, ax2
 
-        if hasattr(self.model, 'evals_result'):
-            fig, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)
-        else:
-            fig, ax1 = plt.subplots(1, 1, figsize=figsize)
+        if ax1 is None:
+            if hasattr(self.model, 'val_result'):
+                _, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)
+            else:
+                _, ax1 = plt.subplots(1, 1, figsize=figsize)
 
         tmpdf = self.results['summary'].sort_values(by='tid', ascending=True)
 
@@ -1051,7 +1237,7 @@ class hgboost:
         ax1.legend()
         if ylim is not None: ax1.set_ylim(ylim)
 
-        if hasattr(self.model, 'evals_result'):
+        if hasattr(self.model, 'val_result'):
             eval_metric = [*self.model.evals_result()['validation_0'].keys()][0]
             ax2.plot([*self.model.evals_result()['validation_0'].values()][0], label='Train error')
             ax2.plot([*self.model.evals_result()['validation_1'].values()][0], label='Test error')
@@ -1063,7 +1249,81 @@ class hgboost:
         if return_ax:
             return ax1, ax2
 
+# %% Plot summary
+# def plot_summary(model, ylim=None, figsize=(15, 10), return_ax=False, method=None, ax1=None, ax2=None, verbose=3):
+#     """Plot the summary results.
 
+#     Parameters
+#     ----------
+#     ylim : tuple
+#         Set the y-limit. In case of auc it can be: (0.5, 1)
+#     figsize: tuple, default (25,25)
+#         Figure size, (height, width)
+
+#     Returns
+#     -------
+#     ax : object
+#         Figure axis.
+
+#     """
+
+#     if method is None:
+#         method = model.method
+
+#     if ('ensemble' in method):
+#         if verbose>=2: print('[hgboost] >Warning: No plot for ensemble is possible yet. <return>')
+#         return None, None
+#     if (not hasattr(model, 'model')):
+#         print('[hgboost] >No model found. Hint: fit a model first using xgboost, catboost or lightboost <return>')
+#         return None, None
+
+#     if ax1 is None:
+#         if hasattr(model, 'val_result'):
+#             _, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)
+#         else:
+#             _, ax1 = plt.subplots(1, 1, figsize=figsize)
+
+#     tmpdf = model.results['summary'].sort_values(by='tid', ascending=True)
+
+#     # Plot results with testsize
+#     idx = np.where(tmpdf['best'].values)[0]
+#     ax1.hlines(tmpdf['loss'].iloc[idx], 0, tmpdf['loss'].shape[0], colors='g', linestyles='dashed', label='Best (wihtout cv)')
+#     ax1.vlines(idx, tmpdf['loss'].min(), tmpdf['loss'].iloc[idx], colors='g', linestyles='dashed')
+#     best_loss = tmpdf['loss'].iloc[idx]
+#     title = ('%s (%s: %.3g)' %(method, model.eval_metric, best_loss))
+
+#     # Plot results with cv
+#     if model.cv is not None:
+#         ax1.errorbar(tmpdf['tid'], tmpdf['loss_mean'], tmpdf['loss_std'], marker='s', mfc='red', label=str(model.cv) + '-fold cv for top ' + str(model.top_cv_evals) + ' models')
+#         idx = np.where(tmpdf['best_cv'].values)[0]
+#         ax1.hlines(tmpdf['loss_mean'].iloc[idx], 0, tmpdf['loss_mean'].shape[0], colors='r', linestyles='dotted', label='Best (' + str(model.cv) + '-fold cv)')
+#         ax1.vlines(idx, tmpdf['loss'].min(), tmpdf['loss_mean'].iloc[idx], colors='r', linestyles='dashed')
+#         best_loss = tmpdf['loss_mean'].iloc[idx]
+#         title = ('%s (%.0d-fold cv mean %s: %.3g)' %(method, model.cv, model.eval_metric, best_loss))
+#         ax1.set_xlabel('Model number')
+
+#     # Plot all other evalution results on the single test-set
+#     ax1.scatter(tmpdf['tid'].values, tmpdf['loss'].values, s=10, label='All models')
+
+#     # Set labels
+#     ax1.set_title(title)
+#     ax1.set_ylabel(model.eval_metric)
+#     ax1.grid(True)
+#     ax1.legend()
+#     if ylim is not None: ax1.set_ylim(ylim)
+
+#     if hasattr(model.model, 'val_result'):
+#         eval_metric = [*model.model.evals_result()['validation_0'].keys()][0]
+#         ax2.plot([*model.model.evals_result()['validation_0'].values()][0], label='Train error')
+#         ax2.plot([*model.model.evals_result()['validation_1'].values()][0], label='Test error')
+#         ax2.set_ylabel(eval_metric)
+#         ax2.set_title(method)
+#         ax2.grid(True)
+#         ax2.legend()
+
+#     if return_ax:
+#         return ax1, ax2
+    
 # %% Import example dataset from github.
 def import_example(data='titanic', url=None, sep=',', verbose=3):
     """Import example dataset from github source.
